import java.lang.*;
import java.util.*;

/** 
 * class Action
 *
 * This class implements a STRIPS action which consists of
 * the fields:
 *
 * name - a descriptive name for the action.  This name must
 *   correspond to the name of an implemented Actor.  It is used
 *   as a key into the ActorTable hash to look up a corresponding
 *   actor.
 * preconditionList - a list of facts which must be true to allow
 *   this action to continue.
 * addList - a list of predicates which will become part of the
 *   agent's state if the action is successful.
 * deleteList - a list of predicates which will be removed from the
 *   agent's state if the action is successful.
 *
 * Action objects are instantiated in the function initializeActionList
 * in StripsBrain.
 *
 * history:
 * 2000.11.29 alokem creation
 */
class Action extends Object
{
  public String name;
  public Vector preconditionList;
  public Vector addList;
  public Vector deleteList;

  /**
   * Constructor
   *
   * String n - name
   * String p - preconditionList (space-delimited list)
   * String a - addList (space-delimited list)
   * String d - deleteList (space delimited list)
   *
   * preconditionList lists the predicates that must be
   * true for this Action to be viable.  The add and delete
   * lists describe the predicates that will be added or
   * removed from the WorldView if the action is carried
   * out.
   *
   * history:
   * 2000.11.30 alokem creation
   */
  
  Action(String n, String p, String a, String d){
    StringTokenizer theString;

    name = new String(n);

    // convert space-delimited list of tokens into a Vector
    theString = new StringTokenizer(p);
    preconditionList = new Vector();
    while (theString.hasMoreTokens())
      preconditionList.add(theString.nextToken());
    
    theString = new StringTokenizer(a);
    addList = new Vector();
    while (theString.hasMoreTokens())
      addList.add(theString.nextToken());

    theString = new StringTokenizer(d);
    deleteList = new Vector();
    while (theString.hasMoreTokens())
      deleteList.add(theString.nextToken());
  }

  /**
   * toString
   *
   * Routine to pretty-print an Action object
   *
   * history:
   * 2000.11.30 alokem creation
   */
  public String toString() {
    return ("action: " + name +
      "\n  preconditionList: " + preconditionList +
      "\n  addList:          " + addList +
      "\n  deleteList:       " + deleteList);
  }

  /**
   * boolean fulfillsGoal(String goal)
   *
   * A function returning a boolean value which indicates 
   * whether the current action object contains the specified
   * goal in its addList.  In other words: does this action
   * accomplish the goal?
   *
   * history:
   * 2000.11.30 alokem creation
   */
  public boolean fulfillsGoal(String goal) {
    Iterator addListIter = addList.iterator();
    boolean retval = false;

    // iterate over strings in addList and compare with the goal
    while (addListIter.hasNext() && (retval == false)) {
      if (goal.compareTo((String)addListIter.next()) == 0) {
        retval = true;
      }
    }

    return retval;
  } // fulfillsGoal

} // Action 