import java.lang.*;
import java.util.*;

/** 
 * interface Actor
 *
 * This interface is used to define classes (also defined in
 * this file) that are capable of achieving a particular 
 * goal.  This approach was taken in order to allow methods to 
 * be looked up using the ActorTable hash.
 * This is the Java approach to function pointers.
 * Since methods are not objects, instead we wrap the
 * sense() method in a Sensor class and pass this around
 * instead.
 * 
 * Objects instantiating the classes here are placed in a 
 * hashtable which is keyed by a String (like "score-goal")
 * which describes what action the Actor will perform.  
 * If you add a new Actor object:
 * - add a line to the initializeActorTable in StripsBrain.java
 *     to instantiate a single object of the class
 * - when creating Actions in the function initializeActionList
 *     in StripsBrain the name property for the Actor
 *     can be used in the name field.
 * - there should be one entry in the ActionList for each
 *     Actor object.
 * 
 * When a plan has been created (i.e. a Vector of strings)
 * the ActorTable will be used to look up the corresponding
 * Actor object and the execute() method will be called.
 *
 * String name - this encodes an action such as "score-goal".  
 * boolean execute - function able to achieve the specified
 *   action (given that the preconditions are met)
 *
 * history:
 * 2000.11.29 alokem creation
 */

interface Actor
{
  public String getName();
  public boolean execute(Memory m, SendCommand actuator, char side);
} // Actor 

// actors

/**
 * actor actorScoreGoal
 *
 * goal: score-goal
 *
 * kicks the ball into opponent's goal.
 */

class actorScoreGoal implements Actor
{ 
  public static final String name = "score-goal";

  public String getName(){
    return name;
  }

  public boolean execute(Memory m, SendCommand actuator, char side){
    ObjectInfo object;
    
    System.out.println("executing " + name);
    object = m.getObject((side == 'l') ? "goal r" : "goal l");
    if (object != null) {
      actuator.kick(100, object.m_direction);
      return true;
    } else {
      return false;
    }
  }
}; // actorScoreGoal

/**
 * actor actorFindBall
 *
 * goal: find-ball
 *
 * Turn until the ball comes into view
 */

class actorFindBall implements Actor
{ 
  public static final String name = "find-ball";

  public String getName(){
    return name;
  }

  public boolean execute(Memory m, SendCommand actuator, char side){
    System.out.println("executing " + name);
    ObjectInfo object;

    while ((object = m.getObject("ball")) == null)
    {
      actuator.turn(40);
      m.waitForNewInfo();
      try {
      	Thread.sleep(2 * SoccerParams.simulator_step);
      } catch (Exception e) {}
    }

    return true;    
  }
}; // actorFindBall

/**
 * actor actorRunToBall
 *
 * goal: run-to-ball
 *
 * Try to get control of ball. 
 */
class actorRunToBall implements Actor
{ 
  public static final String name = "run-to-ball";

  public String getName(){
    return name;
  }

  public boolean execute(Memory m, SendCommand actuator, char side){
    System.out.println("executing " + name);
    ObjectInfo object = m.getObject("ball");

    // if we cannot see the object then something is wrong
    // replan
    if (object == null)
      return false;

    if (object.m_distance > 1.5)
    {
      if (object.m_direction != 0){
        actuator.turn(object.m_direction);
        // pause to avoid back-to-back commands
        try {
          Thread.sleep(2 * SoccerParams.simulator_step);
        } catch (Exception e) {}
      }
      actuator.dash(10 * object.m_distance);
    }

    return true;
  }
}; // actorRunToBall

/**
 * actor actorFindGoal
 *
 * goal: find-goal
 *
 * Look around until goal is visible.
 */


class actorFindGoal implements Actor
{ 
  public static final String name = "find-goal";

  public String getName(){
    return name;
  }

  public boolean execute(Memory m, SendCommand actuator, char side){
    System.out.println("executing " + name);
    ObjectInfo object;
    
    while ((object = m.getObject((side == 'l') ? "goal r" : "goal l")) == null)
    {
      actuator.turn(40);
      m.waitForNewInfo();
      try {
        Thread.sleep(2 * SoccerParams.simulator_step);
      } catch (Exception e) {}
    }

    return true;
  }
}; // actorFindGoal