import java.lang.*;
import java.util.*;

/** 
 * interface Sensor
 *
 * This interface is used to define classes (also defined in
 * this file) that are capable of figuring out whether a given 
 * predicate is true or not.  This approach was taken in order
 * to allow methods to be looked up using the SensorTable hash.
 * This is the Java approach to function pointers.
 * Since methods are not objects, instead we wrap the
 * sense() method in a Sensor class and pass this around
 * instead.
 * 
 * Objects instantiating the classes here are placed in a 
 * hashtable which is keyed by a String (like "can-see-ball")
 * which describes what the Sensor can sense.  If you add a 
 * new Sensor object:
 * - add a line to the initializeSensorTable in StripsBrain.java
 *     to instantiate a single object of the class
 * - when creating Actions (in initializeActionList in StripsBrain) 
 *     the name property for the Sensor can be used in the precondition, 
 *     add or delete lists.
 * 
 * The WorldView object will call 
 * the sense() method of classes that implement the Sensor
 * interface to find out whether the corresponding predicate
 * is true.
 *
 * String name - this encodes a predicate such as "have-ball".  
 * boolean sense - function able to decide using the
 *   agent's memory whether the predicate is true or false
 *
 * history:
 * 2000.11.29 alokem creation
 */
interface Sensor 
{
  public String getName();
  public boolean sense(Memory m, char side);
}

// sensors

/**
 * sensor sensorHaveBall
 *
 * predicate: have-ball
 *
 * Is the ball visible and close by?
 */

class sensorHaveBall implements Sensor
{
  public static final String name = "have-ball";

  public String getName(){
    return name;
  }

  public boolean sense(Memory m, char side)
  {
    boolean retval = false;
    ObjectInfo object = m.getObject("ball");

    if ((object != null) && (object.m_distance < 1.5))
      retval = true;
    
    return retval;
  }
}; // sensorHaveBall

/**
 * sensor sensorCanSeeGoal
 *
 * predicate: can-see-goal
 *
 * opponent's goal is in the visual database
 */

class sensorCanSeeGoal implements Sensor
{
  public static final String name = "can-see-goal";

  public String getName(){
    return name;
  }

  public boolean sense(Memory m, char side)
  {
    ObjectInfo object = m.getObject((side == 'l') ? "goal r" : "goal l");
    return !(object == null);  
  }
}; // sensorCanSeeGoal

/**
 * sensor sensorCanSeeBall
 *
 * predicate: can-see-ball
 *
 * Is the ball visible?
 */

class sensorCanSeeBall implements Sensor
{
  public static final String name = "can-see-ball";

  public String getName(){
    return name;
  }

  public boolean sense(Memory m, char side)
  {
    ObjectInfo object = m.getObject("ball");
    return !(object == null);
  }
}; // sensorCanSeeBall

/**
 * sensor sensorBallInNet
 *
 * predicate: ball-in-net
 *
 * Is the ball in the net.  This is always set to FALSE
 * so that the planner will attempt to repeat the 
 * behaviour.
 */

class sensorBallInNet implements Sensor
{
  public static final String name = "ball-in-net";

  public String getName(){
    return name;
  }

  public boolean sense(Memory m, char side)
  {
    return false;
  }
}; // sensorBallInNet