import java.lang.*;
import java.util.*;

/**
 * class StripsBrain
 *
 * This class implements a brain for a Krislet which uses
 * STRIPS planning to plan its actions for each cycle.  
 * It is adapted from the Brain class by Krzysztof Langer.
 *
 * history:
 * 2000.11.30 alokem creation
 */

class StripsBrain extends Thread implements SensorInput
{

  /**
   * StripsBrain constructor
   *
   * This constructor is a copy of the Brain constructor by
   * Krzysztof Langner.
   *
   * history:
   * 2000.11.30 alokem creation from Krislet Brain
   */
  public StripsBrain(SendCommand myBody, String team,
                     char side, int number, String playMode)
  {
     m_timeOver = false;

     // an object which can be used to make my plans come to life
     m_actuator = myBody;

     // an object which represents the environment for my sensors
     // to work on
     m_memory = new Memory();

     // what side is our goal on
     m_side = side;

     // let's play ball!
     start();
  } // StripsBrain

  /**
   * Hashtable initializeActorTable
   *
   * Add new Actor functions here!
   *
   * This function instantiates the Actor classes defined in 
   * Actor.java and adds them into the actorTable hash.  This
   * allows the Actors to be manipulated as strings (i.e. their
   * name) just like the Actions and Sensors.
   *
   * When the Actor's execute method must be called, we feed
   * the String into the ActorTable hash to get a handle on
   * the Actor whose execute() method we can then invoke.  A Java version
   * of function pointers (kind of like functors).
   *
   * history:
   * 2000.11.30 alokem creation
   */
  private static Hashtable initializeActorTable()
  {
    Hashtable actorTable = new Hashtable();
    Actor     a;

    a = new actorScoreGoal();    actorTable.put(a.getName(), a);
    a = new actorFindBall();     actorTable.put(a.getName(), a);
    a = new actorRunToBall();    actorTable.put(a.getName(), a);
    a = new actorFindGoal();     actorTable.put(a.getName(), a);

    return actorTable;
  }

  /**
   * Hashtable initializeSensorTable
   *
   * Add new Sensors here!
   *
   * Just like the initializeActorTable function, this one instantiates
   * Sensor classes defined in Sensor.java and adds them into a hash
   * which can be used to get a Sensor object, given the Sensor's name,
   * whose sense() method can then be invoked.
   *
   * history:
   * 2000.11.30 alokem creation
   */
  private static Hashtable initializeSensorTable()
  {
    Hashtable sensorTable = new Hashtable();
    Sensor s;

    s = new sensorHaveBall();   sensorTable.put(s.getName(), s);
    s = new sensorCanSeeGoal(); sensorTable.put(s.getName(), s);
    s = new sensorCanSeeBall(); sensorTable.put(s.getName(), s);
    s = new sensorBallInNet();  sensorTable.put(s.getName(), s);

    return sensorTable;
  }

  /**
   * Vector initializeActionList
   *
   * Add new Actions here!
   *
   * Creates the actionList which is iterated over each time a
   * goal needs to be achieved to see which actions will achieve
   * the goal.  This is where the Action objects are instantiated.
   *
   * The name must be a String that corresponds to the name string
   *   of an Actor object.  The name is used as the key to find an
   *   instance of the appropriate Actor object from the ActorTable.
   * The preconditionList, addList and deleteList must be Strings 
   *   containing a list of space delimited words describing predicates
   *   that correspond to the name of an implemented Sensor.  The
   *   individual tokens are used as a key into the SensorTable hash to
   *   look up a corresponding Sensor object which can sense the
   *   predicate's value.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  private static Vector initializeActionList()
  {
    Vector actionList = new Vector();

    actionList.add(new Action("score-goal",             // name
                              "have-ball can-see-goal", // pre
                              "ball-in-net",            // add
                              "have-ball"));            // delete

    actionList.add(new Action("find-goal",              // name
                              "!can-see-goal",          // pre
                              "can-see-goal",           // add
                              ""));                     // delete

    actionList.add(new Action("run-to-ball",            // name
                              "can-see-ball",           // pre
                              "have-ball",              // add
                              ""));                     // delete
  
    actionList.add(new Action("find-ball",              // name
                              "",                       // pre
                              "can-see-ball",           // add
                              ""));                     // delete
    
    return actionList;
  }

  /**
   * executePlan(plan, actorTable, memory, body, side)
   *
   * Vector plan - a vector of strings, each string represents
   *   an Actor and is converted using...
   * Hashtable actorTable - the actorTable hash, initialized
   *   above
   * Memory memory - representation of the environment used by
   *   Actor functions
   * SendCommand body - an interface allowing the Actors to
   *   control the "body" of our soccer player
   * char side - 'l' or 'r', our side during the first half
   *   (note: orientations are changed in second half so we still shoot
   *   on the same goal)
   *
   * Iterate over the plan, convert the strings into Actors 
   * and then execute() them.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  private static void executePlan(Vector plan, Hashtable actorTable, 
    Memory memory, SendCommand body, char side)
  {
    Iterator planIter = plan.iterator();

    System.out.println("okay, here's the plan: " + plan);

    while (planIter.hasNext())
    {
      String actorName = (String)planIter.next();

      // string -> actor conversion
      Actor a = (Actor)actorTable.get(actorName);

      // if there was a failure in the plan, exit immediately
      // and re-plan
      if (a.execute(memory, body, side) == false) {
        System.out.println("could not execute " + actorName);
        return;
      }
    }
  }

  /**
   * boolean doStripsPlanning
   *
   * The heart of the Stripslet is the STRIPS algorithm
   * which is implemented here.  The algorithm is described
   * inline.
   *
   * in:
   * Vector goals - Vector of strings which should be resolved
   *   by planning.
   * WorldView worldView - the state of the agent's environment
   *   it is not completely specified, instead we fill it in
   *   on demand
   * Vector actions - the list of all possible actions.  By
   *   iterating through this, we can hopefully find an action
   *   which will achieve our goal.
   * 
   * out:
   * Vector plan - a Vector of Strings which represent Actors that
   *   cause the goal to be achieved.  Feed it to executePlan in order 
   *   to translate it into actions on the field.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  private static boolean
  doStripsPlanning(Vector goals, WorldView worldView, 
    Vector actions, Vector plan)
  {
    Iterator goalList = goals.iterator();

    // look at each goal in the goal list
    boolean allGoalsAchieved = true;
    while (goalList.hasNext() && (allGoalsAchieved == true))
    {
      // check if goal has already been achieved
      String  goal         = (String)goalList.next();
      boolean goalAchieved = worldView.evaluate(goal);

      // if yes then
      //   go on to next goal
      if (goalAchieved)
        continue;

      // else look at each action in the action list
      Iterator actionList = actions.iterator();
      boolean foundAction = false;
      while (actionList.hasNext() && (foundAction == false))
      {
        Action action = (Action)actionList.next();
        // will doing this action achieve the goal?
        // if so attempt to fulfill its preconditions
        if (action.fulfillsGoal(goal) &&
            doStripsPlanning(action.preconditionList, worldView, actions, plan))
        {
          System.out.println("can fullfill " + goal + " with action " + action.name);
          // if preconditions were fulfilled then add action to plan
          // apply adds and deletes to worldView
          plan.add(action.name);
          worldView.add(action.addList);
          worldView.delete(action.deleteList);
          // we found a good action, so don't bother continuing
          // through action list
          foundAction = true;
        }
      } // looking through action list

      // if we didn't find an action that both fulfilled the goal and
      // could be planned, we cannot achieve all our goals
      if (foundAction == false)
        allGoalsAchieved = false;

    } // looking through goals

    return allGoalsAchieved;
  } // doStripsPlanning

  /**
   * run
   *
   * The main loop of the thinking thread (as opposed to the
   * sensing thread which fills in the Memory object).  We start
   * out by initializing our hashtables (actor and sensorTable)
   * and the ActionList.  
   *
   * In the main loop we:
   *   - refresh our worldview
   *   - designate a goal
   *   - generate a plan to achieve the goal
   *   - if generated successfully execute the plan
   *
   * history:
   * 2000.11.30 alokem creation
   */


  public void run()
  {
    // mapping of strings to actors
    Hashtable actorTable = initializeActorTable();

    // mapping of strings to sensors
    Hashtable sensorTable = initializeSensorTable();

    // list of possible STRIPS actions
    Vector actionList = initializeActionList();
    
    // planning cycle
    //   re-initialize state of world
    //   decide what our goals are this round
    //   do planning
    //   execute plan
    while (!m_timeOver)
    {
      // we have a new worldview after every cycle
      WorldView worldView   = new WorldView(sensorTable, m_memory, m_side);
      Vector    goalList    = new Vector();
      Vector    plan        = new Vector();

      // the goal we are trying to achieve, ideally this could be the
      // output of another function which uses the memory to assess which
      // goal would be most fruitful
      goalList.add("ball-in-net");

      // attempt to generate a plan for achieving the goals in the 
      // goalList
      if (doStripsPlanning(goalList, worldView, actionList, plan)) {
        executePlan(plan, actorTable, m_memory, m_actuator, m_side);
      }

			// sleep one step to ensure that we will not send
			// two commands in one cycle.
			try {
				Thread.sleep(2*SoccerParams.simulator_step);
			} catch(Exception e) {}

    } // end planning-cycle

  } // run

  // implementation of SensorInput interface
  // These are borrowed from Krzysztof Langer's Krislet Brain.

  // caches any visual info retrieved by the main thread
  // in this brain's memory
	public void see(VisualInfo info) {
		m_memory.store(info);
	}

  // not implemented
	public void hear(int time, int direction, String message)
	{
	}

  // not implemented
	public void hear(int time, String message)
	{	
	}

  // actuator required to execute actions
	private SendCommand	m_actuator;

  // memory base used by WorldView to calculate the
  // truth of predicates
	private Memory			m_memory;

  // a character [l, r] which can be used to deduce which goal
  // object to shoot on 
	private char				m_side;

  // indicates when the game is over
	volatile private boolean		m_timeOver;

} // StripsBrain