import java.lang.*;
import java.util.*;

/**
 * class WorldView
 *
 * This class represents the state of the agent's environment.
 *
 * It works as a kind of cache for values of predicates.
 * The two main components are the factStates and sensorTable
 * hashes.  
 * 
 * The sensorTable hash (initialized in StripsBrain)
 * is a mapping of predicates like "have-ball" to Sensor
 * objects which are able to find the value of these predicates.
 *
 * The factStates hash allows sensor queries to be cached.
 * The hash is initially empty, but fills in with truth values
 * as successive queries are performed.  When the planning process
 * decides on an action, the add and delete lists are also
 * applied to the worldView.
 *
 * The worldView should be re-initialized after every planning
 * cycle in order to reflect the latest state of the environment
 * (represented by the memory object).
 *
 * history:
 * 2000.11.30 alokem creation
 */

class WorldView extends Object
{
  private Hashtable factStates;
  private Hashtable sensorTable;
  private Memory memory;
  private char side;

  /**
   * Constructor 
   *
   * args:
   *   Hashtable s - sensorTable
   *   Memory    m - agent's memory (represent environment)
   *   char      theSide - the side of the agent's goal
   *
   * history:
   * 2000.11.30 alokem creation
   */
  WorldView(Hashtable s, Memory m, char theSide)
  {
    sensorTable = s;
    memory = m;
    side = theSide;
    factStates = new Hashtable();
  }

  /**
   * evaluate
   *
   * Returns a truth-value indicating the state of the
   * predicate (factName).  If value is contained in factStates
   * hash, return it immediately, otherwise find this
   * predicate's Sensor and invoke its sense() method to
   * find the predicate's value.  Store the returned value in
   * factStates.
   *
   * args:
   *   String factName - String representing a predicate.
   *     this name should correspond to a Sensor object.
   *
   * out:
   *   boolean - state of predicate.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  public boolean evaluate(String factName)
  {
    boolean negationFlag;
    Boolean state;
    
    // if first character is ! chop it off and set negationFlag
    // all operations done on unnegated state and then flipped
    // in the return statement
    negationFlag = (factName.charAt(0) == '!');
    if (negationFlag == true)
      factName = factName.substring(1);

    // check if the state's value has already been seen
    state = (Boolean)factStates.get(factName);

    // fact was not found in hash, so go out to sensors and
    // find out whether fact is true or not
    if (state == null)
    {
      Sensor s = (Sensor)sensorTable.get(factName);
      state = new Boolean(s.sense(memory, side));
      // store result in factStates
      factStates.put(factName, state);
      System.out.println("sensed " + factName + " was " + state);
    }

    // flip the result if negationFlag == true
    return (negationFlag ^ state.booleanValue());
  }

  /**
   * add(Vector)
   *
   * Used to add a set of facts into the WorldView.  Called from 
   * planning to reflect the add, delete lists specified in an
   * Action.
   * An added value will override all Sensor information.
   *
   * args:
   *   Vector factName - array of Strings, each element representing
   *     a predicate to add.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  public void add(Vector factNames)
  {
    Iterator facts = factNames.iterator();

    while (facts.hasNext())
      add((String)facts.next());
  } // add(Vector)

  /**
   * add(String)
   *
   * Used to add a single fact into the WorldView.  Note that if
   * predicate value is preceded by a ! (i.e. !have-ball) then the
   * value added in the hashTable will be the complement (i.e false)
   * because there are no complemented predicates in the worldView.
   *
   * args:
   *   Vector factName - array of Strings, each element representing
   *     a predicate to add.
   *
   * history:
   * 2000.11.30 alokem creation
   */

  public void add(String factName)
  {
    boolean negationFlag;

    // if first character is ! chop it off and set negationFlag
    negationFlag = (factName.charAt(0) == '!');
    if (negationFlag == true)
      factName = factName.substring(1);

    // adding (!fact = true) is the same as (fact = false)
    // and vice versa
    factStates.put(factName, new Boolean(!negationFlag));
  } // add(String)

  /**
   * delete(Vector)
   *
   * Similar to add(Vector) but to remove (set predicates to false)
   * facts from worldView.  
   *
   * history:
   * 2000.11.30 alokem
   */
  
  public void delete(Vector factNames)
  {
    Iterator facts = factNames.iterator();

    while (facts.hasNext())
      delete((String)facts.next());
  } // delete(Vector)

  /**
   * delete(String)
   *
   * Similar to add(String) but to remove (set predicates to false)
   * facts from WorldView.  The opposite behaviour from add applies
   * for predicates preceded by a !.
   *
   * history:
   * 2000.11.30 alokem
   */

  public void delete(String factName)
  {
    boolean negationFlag;

    // if first character is ! chop it off and set negationFlag
    negationFlag = (factName.charAt(0) == '!');
    if (negationFlag == true)
      factName = factName.substring(1);

    // deleting (!fact = true) is the same as setting (fact = false)
    // and vice versa
    factStates.put(factName, new Boolean(negationFlag));
  } // delete(String)

  /**
   * toString
   *
   * Function that can be used to pretty-print this data structure.
   *
   * history:
   * 2000.11.30 alokem creation
   */
  public String toString() {
    return ("worldview: \n  sensorTable: " + sensorTable + "\n  factStates: " + factStates);
  }
} // WorldView